// cronwrapper
// Executes a command and gatheres the output of stdout and stderr.
// If the return code of the command is not zero the gathered output will
// be written to stdout which will generate a mail in cron context.

#include <config.h>
#include <sys/queue.h>
#include <sys/wait.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef INFTIM
#define INFTIM (-1)
#endif

#ifdef USE_BSD_H
#include <bsd/string.h>
#endif

int child_status;

struct buffer {
	TAILQ_ENTRY(buffer) buffers;
	size_t size;
	char   data[1];
};

TAILQ_HEAD(listhead, buffer) output;


struct buffer *
buffer_new(size_t size) {
	struct buffer *b = malloc(sizeof(struct buffer) + size);
	if (b == NULL)
		err(1, "malloc");
	b->size = size;
	b->data[0] = '\0';
	return b;
}


void
reaper(int signal)
{
	int save_errno = errno;

	wait(&child_status);

	errno = save_errno;
}


int
execute(char **argv)
{
	pid_t pid;

	int p_err[2];
	int p_out[2];
	int p_in[2];

	if (pipe(p_err) == -1)
		err(1, "pipe");
	if (pipe(p_out) == -1)
		err(1, "pipe");
	if (pipe(p_in) == -1)
		err(1, "pipe");

	switch (pid = fork()) {
	case -1:
		err(1, "fork");
		break;
	case 0:
		/* Child process */
		signal(SIGCHLD, SIG_DFL);

		close(p_err[0]);
		close(p_out[0]);
		close(p_in[0]);

		if (dup2(p_err[1], STDERR_FILENO) == -1)
			_exit(1);
		if (dup2(p_out[1], STDOUT_FILENO) == -1)
			_exit(1);
		if (dup2(p_in[1], STDIN_FILENO) == -1)
			_exit(1);

		close(p_err[1]);
		close(p_out[1]);
		close(p_in[1]);

		execvp(argv[0], argv);
		/* fall through on error */
		_exit(1);
		break;
	}
	TAILQ_INIT(&output);

	close(p_out[1]);
	close(p_err[1]);
	close(p_in[1]);

	fcntl(p_out[0], F_SETFD, FD_CLOEXEC);
	fcntl(p_err[0], F_SETFD, FD_CLOEXEC);
	fcntl(p_in[0], F_SETFD, FD_CLOEXEC);

	struct pollfd pfd[2] = {
		{ .fd = p_out[0], .events = POLLIN },
		{ .fd = p_err[0], .events = POLLIN },
	};

	int nfds = 2;
	struct pollfd *p = pfd;

	int nready;
	while ((nready = poll(p, nfds, INFTIM)) > 0) {
		if (nready == 0)
			break;

		if (pfd[0].revents & (POLLNVAL|POLLHUP)) {
			break;
		}
		if (pfd[1].revents & (POLLNVAL|POLLHUP)) {
			break;
		}
		if (pfd[0].revents & (POLLERR)) {
			err(1, "poll");
		}
		if (pfd[1].revents & (POLLERR)) {
			err(1, "poll");
		}
		ssize_t n = 0;
		char buf[1024];
		if (pfd[0].revents & (POLLIN)) {
			if ((n = read(pfd[0].fd, buf, sizeof(buf))) == -1)
				err(1, "read");
			struct buffer *b = buffer_new((size_t )n);
			strlcpy(b->data, buf, n);
			TAILQ_INSERT_TAIL(&output, b, buffers);
		}
		if (pfd[1].revents & POLLIN) {
			if ((n = read(pfd[1].fd, buf, sizeof(buf))) == -1)
				err(1, "read");
			struct buffer *b = buffer_new((size_t )n);
			strlcpy(b->data, buf, n);
			TAILQ_INSERT_TAIL(&output, b, buffers);
		}

	}

	waitpid(pid, &child_status, 0);
	return WEXITSTATUS(child_status);
}


int
main(int argc, char **argv)
{
	if (argc < 2)
		errx(1, "usage: %s cmd [args]", argv[0]);

	signal(SIGCHLD, &reaper);
	int result = execute(&argv[1]);
	if (result != 0) {
		struct buffer *b;
		TAILQ_FOREACH(b, &output, buffers) {
			if (write(STDOUT_FILENO, b->data, b->size) != b->size)
				err(1, "write");
		}
		const char *nl = "\n";
		const size_t nllen = strlen("\n");
		if (write(STDOUT_FILENO, nl, nllen) != nllen)
			err(1, "write");
	}
	return result;
}
